@@ -4,6 +4,7 @@
     begin                : Fri Dec 10 16:02:01 MSK 1999
     copyright            : (C) 1999 by Yury Lebedev
     email                : yurylebedev@mail.ru
+    home page            : http://devel-home.kde.org/~ksqlplus
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,10 +19,10 @@
 // include files for QT
 #include <qdir.h>
 #include <qfile.h>
-#include <qstrlist.h>
 #include <qprinter.h>
 #include <qpainter.h>
 #include <qsplitter.h>
+#include <qclipboard.h>
 
 // include files for KDE
 #include <kiconloader.h>
@@ -29,15 +30,17 @@
 #include <kfiledialog.h>
 #include <kfontdialog.h>
 #include <kkeydialog.h>
+#include <kurl.h>
 
-// application specific includes
+// application-specific includes
 #include "ksqlplus.h"
 #include "klogindlg.h"
 #include "khistorydlg.h"
+#include "ksqlplusdlg.h"
 #include "resource.h"
 
 
-KSqlPlusApp::KSqlPlusApp()
+KSqlPlusApp::KSqlPlusApp() : KTMainWindow()
 {
   // create application configuration
   config = kapp->getConfig();
@@ -58,6 +61,9 @@ KSqlPlusApp::KSqlPlusApp()
   // init command history
   commandHistory.setAutoDelete(true);
 
+  // init command queue for SQL*Plus
+  sqlplusCommandQueue.setAutoDelete(true);
+
   // disable menu and toolbar items at startup
   disableCommand(ID_FILE_PRINT);
   disableCommand(ID_FILE_PRINT_OUTPUT);
@@ -67,6 +73,21 @@ KSqlPlusApp::KSqlPlusApp()
   disableCommand(ID_COMMAND_PREV);
   disableCommand(ID_COMMAND_NEXT);
 
+  // setup watcher for clipboard
+  QClipboard* clip = QApplication::clipboard();
+  connect(clip, SIGNAL(dataChanged()), SLOT(slotClipboardChanged()));
+  // check the current clipboard content
+  if( clip->text() == 0 )
+    disableCommand(ID_EDIT_PASTE);
+
+  // setup drag-n-drop
+#if KDE_VERSION_MAJOR == 1
+  // KDE1 drag-n-drop
+  KDNDDropZone* kde1DropZone = new KDNDDropZone(inputView, DndURL);
+  connect(kde1DropZone, SIGNAL(dropAction(KDNDDropZone*)),
+    SLOT(slotDropActionKDE1(KDNDDropZone*)));
+#endif
+
   // update application caption
   updateCaption();
 
@@ -79,7 +100,7 @@ KSqlPlusApp::~KSqlPlusApp()
   // try to finish SQL*Plus correctly (with implicit commit)
   if( sqlplus.isRunning() )
   {
-    sqlplusExecute("\nEXIT\n", false);
+    sqlplusExecute("EXIT\n", false);
   }
 }
 
@@ -96,9 +117,12 @@ void KSqlPlusApp::initKeyAccel()
   keyAccel = new KAccel(this, "keyAccel");
 
   // create application-specific keys
-  keyAccel->insertItem(i18n("Execute command"), KEY_COMMAND_EXECUTE, "CTRL+Return");
-  keyAccel->insertItem(i18n("Previous command"), KEY_COMMAND_PREV, "ALT+Left");
-  keyAccel->insertItem(i18n("Next command"), KEY_COMMAND_NEXT, "ALT+Right");
+  keyAccel->insertItem(i18n("Execute command"), KEY_COMMAND_EXECUTE,
+    "CTRL+Return");
+  keyAccel->insertItem(i18n("Previous command"), KEY_COMMAND_PREV,
+    "ALT+Left");
+  keyAccel->insertItem(i18n("Next command"), KEY_COMMAND_NEXT,
+    "ALT+Right");
 
   // read default settings
   keyAccel->readSettings(config);
@@ -143,11 +167,9 @@ void KSqlPlusApp::initKeyAccel()
 
 void KSqlPlusApp::initMenuBar()
 {
-  ///////////////////////////////////////////////////////////////////
   // MENUBAR
   recentFilesMenu = new QPopupMenu();
 
-  ///////////////////////////////////////////////////////////////////
   // menuBar entry fileMenu
   fileMenu = new QPopupMenu();
   fileMenu->insertItem(kapp->getMiniIcon(), i18n("New &Window"), ID_FILE_NEW_WINDOW);
@@ -168,7 +190,6 @@ void KSqlPlusApp::initMenuBar()
   fileMenu->insertSeparator();
   fileMenu->insertItem(i18n("E&xit"), ID_FILE_QUIT);
 
-  ///////////////////////////////////////////////////////////////////
   // menuBar entry editMenu
   editMenu = new QPopupMenu();
   editMenu->insertItem(Icon("editcut.xpm"), i18n("Cu&t"), ID_EDIT_CUT);
@@ -181,8 +202,8 @@ void KSqlPlusApp::initMenuBar()
   editMenu->insertItem(i18n("Select &all output"), ID_EDIT_SELECT_ALL_OUTPUT);
   editMenu->insertSeparator();
   editMenu->insertItem(i18n("&Keys..."), ID_EDIT_KEYS);
+  editMenu->insertItem(i18n("&SQL*Plus location..."), ID_EDIT_SQLPLUS);
 
-  ///////////////////////////////////////////////////////////////////
   // menuBar entry viewMenu
   viewMenu = new QPopupMenu();
   viewMenu->setCheckable(true);
@@ -192,7 +213,6 @@ void KSqlPlusApp::initMenuBar()
   viewMenu->insertSeparator();
   viewMenu->insertItem(Icon("text.xpm"), i18n("&Font..."), ID_VIEW_FONT);
 
-  ///////////////////////////////////////////////////////////////////
   // menuBar entry commandView
   commandMenu = new QPopupMenu();
   commandMenu->insertItem(Icon("configure.xpm"), i18n("&Execute"), ID_COMMAND_EXECUTE);
@@ -202,17 +222,15 @@ void KSqlPlusApp::initMenuBar()
   commandMenu->insertItem(Icon("forward.xpm"), i18n("&Next command"), ID_COMMAND_NEXT);
   commandMenu->insertItem(i18n("Command &history..."), ID_COMMAND_HISTORY);
 
-  ///////////////////////////////////////////////////////////////////
   // menuBar entry helpMenu
-  QString about = kapp->appName() + " " + VERSION +
-    i18n("\n\nKDE front-end for ORACLE SQL*Plus.\n"
+  QString about = kapp->appName() + " " VERSION +
+    i18n("\n\nKDE frontend for ORACLE SQL*Plus.\n"
     "Developed by Yury Lebedev (yurylebedev@mail.ru).\n\n"
     "This is free software covered by GPL.\n\n"
     "ORACLE and SQL*Plus are registered marks of ORACLE Corporation\n"
     "and these are commercial software.\n");
   helpMenu = kapp->getHelpMenu(true, about);
 
-  ///////////////////////////////////////////////////////////////////
   // MENUBAR CONFIGURATION
   // insert your popup menus with the according menubar entries in the order
   // they will appear later from left to right
@@ -223,7 +241,6 @@ void KSqlPlusApp::initMenuBar()
   menuBar()->insertSeparator();
   menuBar()->insertItem(i18n("&Help"), helpMenu);
 
-  ///////////////////////////////////////////////////////////////////
   // CONNECT THE MENU SLOTS WITH SIGNALS
   // for execution slots and statusbar messages
 
@@ -286,6 +303,7 @@ void KSqlPlusApp::initToolBar()
 
 void KSqlPlusApp::initStatusBar()
 {
+  // STATUSBAR
   statusBar()->setInsertOrder(KStatusBar::RightToLeft);
   // input view cursor position message
   QString s;
@@ -293,6 +311,13 @@ void KSqlPlusApp::initStatusBar()
   statusBar()->insertItem(s, ID_STATUS_ROW_COL);
   s.sprintf(i18n("Line: %d Column: %d "), 1, 1);
   statusBar()->changeItem(s, ID_STATUS_ROW_COL);
+  // overwrite mode flag
+  statusBar()->insertItem(i18n(" OVR "), ID_STATUS_OVERWRITE);
+  statusBar()->changeItem(i18n(" INS "), ID_STATUS_OVERWRITE);
+  statusBar()->setAlignment(ID_STATUS_OVERWRITE, AlignCenter);
+  // modified flag
+  statusBar()->insertItem("   ", ID_STATUS_MODIFIED);
+  statusBar()->setAlignment(ID_STATUS_MODIFIED, AlignCenter);
   // status message
   statusBar()->insertItem(i18n("Ready"), ID_STATUS_MSG);
 }
@@ -300,17 +325,29 @@ void KSqlPlusApp::initStatusBar()
 void KSqlPlusApp::initView()
 {
   // create main view splitter
-  QSplitter* splitter = new QSplitter(QSplitter::Vertical, this, "viewsSplitter");
+  QSplitter* splitter = new QSplitter(QSplitter::Vertical, this,
+    "viewsSplitter");
 
   // create view for SQL*Plus output
   outputView = new QMultiLineEdit(splitter, "outputView");
   outputView->setReadOnly(true);
+  outputView->setFocusPolicy(NoFocus);
 
   // create view for user input
   inputView = new KEdit(kapp, splitter, "inputView");
+  // allow use tabs inside text area
+  inputView->setFocusPolicy(ClickFocus);
+  // allow autoident mode
+  inputView->setAutoIndentMode(true);
   // connect slot
-  connect(inputView, SIGNAL(CursorPositionChanged()), SLOT(slotInputViewCursorChanged()));
+  connect(inputView, SIGNAL(textChanged()),
+    SLOT(slotInputViewTextChanged()));
+  connect(inputView, SIGNAL(CursorPositionChanged()),
+    SLOT(slotInputViewCursorChanged()));
+  connect(inputView, SIGNAL(toggle_overwrite_signal()),
+    SLOT(slotInputViewOverwriteToggled()));
 
+  // setup the splitter
   splitter->setOpaqueResize(true);
   // set splitter as main view
   setView(splitter);
@@ -324,9 +361,14 @@ void KSqlPlusApp::initSqlplus()
   // init SQL*Plus process
 
   // connect signals to slots
-  connect(&sqlplus, SIGNAL(processExited(KProcess*)), this, SLOT(slotSqlplusExited(KProcess*)));
-  connect(&sqlplus, SIGNAL(receivedStdout(KProcess*,char*,int)), this, SLOT(slotSqlplusStdout(KProcess*,char*,int)));
-  connect(&sqlplus, SIGNAL(receivedStderr(KProcess*,char*,int)), this, SLOT(slotSqlplusStdout(KProcess*,char*,int)));
+  connect(&sqlplus, SIGNAL(processExited(KProcess*)),
+    SLOT(slotSqlplusExited(KProcess*)));
+  connect(&sqlplus, SIGNAL(wroteStdin(KProcess*)),
+    SLOT(slotSqlplusWroteStdin(KProcess*)));
+  connect(&sqlplus, SIGNAL(receivedStdout(KProcess*,char*,int)),
+    SLOT(slotSqlplusStdout(KProcess*,char*,int)));
+  connect(&sqlplus, SIGNAL(receivedStderr(KProcess*,char*,int)),
+    SLOT(slotSqlplusStdout(KProcess*,char*,int)));
 }
 
 void KSqlPlusApp::updateCaption()
@@ -339,7 +381,11 @@ void KSqlPlusApp::updateCaption()
   // add current filename
   QString filename;
   if( isNew() ) filename = i18n("New SQL script");
-  else filename = inputView->getName();
+  else
+  {
+    QFileInfo fi(inputView->getName());
+    filename = fi.fileName();
+  }
   caption += " - " + filename;
   // set main window caption
   setCaption(caption);
@@ -359,12 +405,11 @@ void KSqlPlusApp::appendToOutputView(const char* text)
   outputView->setAutoUpdate(true);
   outputView->repaint();
   // scroll down output view
-  outputView->setCursorPosition(outputView->numLines(), 1);
+  outputView->setCursorPosition(outputView->numLines() - 1, 1);
 }
 
 void KSqlPlusApp::enableCommand(int id)
 {
-  ///////////////////////////////////////////////////////////////////
   // enable menu and toolbar functions by their ID's
   menuBar()->setItemEnabled(id, true);
   for( int i = 0; i < MAX_TOOLBARS; i++ )
@@ -373,7 +418,6 @@ void KSqlPlusApp::enableCommand(int id)
 
 void KSqlPlusApp::disableCommand(int id)
 {
-  ///////////////////////////////////////////////////////////////////
   // disable menu and toolbar functions by their ID's
   menuBar()->setItemEnabled(id, false);
   for( int i = 0; i < MAX_TOOLBARS; i++ )
@@ -383,7 +427,9 @@ void KSqlPlusApp::disableCommand(int id)
 void KSqlPlusApp::saveOptions()
 {
   // save general settings
+  QString prev(config->group());
   config->setGroup("General Options");
+
   config->writeEntry("Geometry", size());
   config->writeEntry("Show Toolbar", toolBar()->isVisible());
   config->writeEntry("Show Statusbar",statusBar()->isVisible());
@@ -392,12 +438,16 @@ void KSqlPlusApp::saveOptions()
   config->writeEntry("InputToolBar Position", (int)toolBar(1)->barPos());
   config->writeEntry("Recent Files", recentFiles);
   config->writeEntry("Font", viewsFont);
+
+  config->setGroup(prev);
+
   // save accelerators
   keyAccel->writeSettings(config);
 }
 
 void KSqlPlusApp::readOptions()
 {
+  QString prev(config->group());
   config->setGroup("General Options");
 
   // bar status settings
@@ -443,34 +493,13 @@ void KSqlPlusApp::readOptions()
   // set restored font for input/output views
   outputView->setFont(viewsFont);
   inputView->setFont(viewsFont);
-}
 
-const QString KSqlPlusApp::readUsername()
-{
-  config->setGroup("ORACLE Options");
-  return config->readEntry("Username");
-}
-
-const QString KSqlPlusApp::readDatabase()
-{
-  config->setGroup("ORACLE Options");
-  return config->readEntry("Database");
-}
-
-void KSqlPlusApp::saveUsernameAndDatabase(const char* username, const char* database)
-{
-  config->setGroup("ORACLE Options");
-  config->writeEntry("Username", username);
-  config->writeEntry("Database", database);
+  config->setGroup(prev);
 }
 
 void KSqlPlusApp::saveProperties(KConfig *cfg)
 {
-  if( isNew() || !inputView->isModified() )
-  {
-    // saving to tempfile not necessary
-  }
-  else
+  if( !isNew() && inputView->isModified() )
   {
     QString filename = inputView->getName();
     cfg->writeEntry("filename", filename);
@@ -514,29 +543,28 @@ bool KSqlPlusApp::isNew()
 
 bool KSqlPlusApp::saveModified()
 {
-  if( inputView->isModified() )
+  if( !inputView->isModified() ) return true;
+
+  int wantSave = KMsgBox::yesNoCancel(this, i18n("Warning"),
+    i18n("The current SQL script has been modified.\nDo you want to save it?"),
+    KMsgBox::QUESTION, i18n("Yes"), i18n("No"), i18n("Cancel"));
+
+  switch( wantSave )
   {
-    int want_save = KMsgBox::yesNoCancel(this, i18n("Warning"),
-      i18n("The current SQL script has been modified.\nDo you want to save it?"),
-      KMsgBox::QUESTION, i18n("Yes"), i18n("No"), i18n("Cancel"));
-    switch( want_save )
-    {
-      case 1:
-        if( isNew() ) slotFileSaveAs();
-        else inputView->doSave();
-        return true;
-      case 2:
-        inputView->toggleModified(false);
-        return true;
-      default:
-        return false;
-    }
+    case 1:
+      slotFileSave();
+      return true;
+    case 2:
+      inputView->toggleModified(false);
+      // update modified flag in statusbar
+      slotInputViewTextChanged();
+      return true;
   }
 
-  return true;
+  return false;
 }
 
-void KSqlPlusApp::addRecentFile(const QString &file)
+void KSqlPlusApp::addRecentFile(const char* file)
 {
   if( recentFiles.find(file) == -1 )
   {
@@ -622,12 +650,54 @@ bool KSqlPlusApp::queryExit()
 // SLOT IMPLEMENTATION
 /////////////////////////////////////////////////////////////////////
 
+#if KDE_VERSION_MAJOR == 1
+void KSqlPlusApp::slotDropActionKDE1(KDNDDropZone* drop)
+{
+  // check modified document
+  if( !saveModified() ) return;
+
+  // get list of URLs
+  QStrList& urls = drop->getURLList();
+  // open list of URLs - first in current window, nexts - in new windows
+  bool first = true;
+  for( const char* url = urls.first(); url; url = urls.next() )
+  {
+    // check for URL's protocol
+    KURL kurl(url);
+    if( kurl.isMalformed() || !kurl.isLocalFile() || kurl.hasSubProtocol() )
+    {
+      QString msg = i18n("Cannot open URL ");
+      msg += url;
+      msg += '\n';
+      msg += kapp->appName();
+      msg += " " VERSION;
+      msg += i18n(" can work with local files only");
+      KMsgBox::message(this, i18n("Error"), msg, KMsgBox::EXCLAMATION);
+      continue;
+    }
+    if( first )
+    {
+      // open first file in current window
+      openDocumentFile(url);
+      first = false;
+    }
+    else
+    {
+       // create new window for next file
+       KSqlPlusApp* win = new KSqlPlusApp();
+       win->show();
+       win->openDocumentFile(url);
+    }
+  }
+}
+#endif
+
 void KSqlPlusApp::slotFileNewWindow()
 {
   slotStatusMsg(i18n("Opening a new application window..."));
 
-  KSqlPlusApp* new_window = new KSqlPlusApp();
-  new_window->show();
+  KSqlPlusApp* win = new KSqlPlusApp();
+  win->show();
 
   slotStatusMsg(i18n("Ready"));
 }
@@ -636,36 +706,20 @@ void KSqlPlusApp::slotFileLogin()
 {
   slotStatusMsg(i18n("Logging on into ORACLE..."));
 
-  // login parameters
-  QString connect;
-  QString username;
-  QString database;
-  bool do_login = false;
-
-  // process Login dialog
-  // (fixed 'core dump' after cancel dialog and close application)
-  {
-    KLoginDlg dlg(this, "loginDlg", readUsername(), readDatabase());
-    if( dlg.exec() )
-    {
-      // OK is pressed
-      do_login = true;
-      connect = dlg.connectString();
-      username = dlg.username();
-      database = dlg.database();
-      // save 'username@database'
-      lastOracleIdentity = dlg.identityString();
-    }
-  }
-
-  if( do_login )
+  // create and show Login dialog (it'll be destroyed automatically)
+  KLoginDlg* loginDlg = new KLoginDlg(this, "loginDlg");
+  if( loginDlg->exec() )
   {
+    // OK is pressed
+    // save new login parameters
+    lastOracleIdentity = loginDlg->identityString();
+    QString remoteHost = loginDlg->remoteHost();
+    // setup SQL*Plus process arguments
+    loginDlg->setupSqlplus(sqlplus);
     // do real login
     if( !sqlplus.isRunning() )
     {
-      // first time logon - pass connect string as argument
-      sqlplus << "sqlplus" << connect;
-      // start SQL*Plus
+      // try to start SQL*Plus
       if( !sqlplus.start(KProcess::NotifyOnExit, KProcess::All) )
       {
         KMsgBox::message(this, i18n("Error"),
@@ -680,18 +734,29 @@ void KSqlPlusApp::slotFileLogin()
     }
     else
     {
-      // next time logon - pass as command "CONNECT"
-      QString cmd("CONNECT ");
-      cmd += connect;
-      cmd += '\n';
-      sqlplusExecute(cmd, true);
-      // and set it as new argument for SQL*Plus process
-      sqlplus.clearArguments();
-      sqlplus << "sqlplus" << connect;
+      // next time logon
+      // check the identity of remote SQL*Pluses
+      if( lastRemoteHost != remoteHost )
+      {
+        KMsgBox::message(this, i18n("Warning"),
+          i18n("You have chosen another host of SQL*Plus.\n"
+          "This instance of SQL*Plus will be restarted now\n"
+          "with new login parameters."),
+          KMsgBox::INFORMATION);
+        sqlplusExecute("EXIT\n", false);
+      }
+      else
+      {
+        // pass connect string as command "CONNECT"
+        QString cmd("CONNECT ");
+        cmd += loginDlg->connectString();
+        cmd += '\n';
+        sqlplusExecute(cmd, true);
+      }
     }
 
-    // write connect parameters to config
-    saveUsernameAndDatabase(username, database);
+    // save host of SQL*Plus
+    lastRemoteHost = remoteHost;
     // update window title
     updateCaption();
   }
@@ -712,11 +777,7 @@ void KSqlPlusApp::slotFileNew()
 {
   slotStatusMsg(i18n("Creating new SQL script..."));
 
-  if( !saveModified() )
-  {
-     // here saving wasn't successful
-  }
-  else
+  if( saveModified() )
   {
     inputView->newFile();
     updateCaption();
@@ -729,15 +790,11 @@ void KSqlPlusApp::slotFileOpen()
 {
   slotStatusMsg(i18n("Opening SQL script..."));
 
-  if( !saveModified() )
-  {
-     // here saving wasn't successful
-  }
-  else
+  if( saveModified() )
   {
     QString file = KFileDialog::getOpenFileName(QDir::currentDirPath(),
-      i18n("*.sql|SQL scripts\n*|All files"), this, i18n("Open File..."));
-
+      i18n("*.sql *.SQL|SQL scripts\n*|All files"), this);
+    // open chosen file by usual way
     if( !file.isEmpty() ) openDocumentFile(file);
   }
 
@@ -748,11 +805,7 @@ void KSqlPlusApp::slotFileOpenRecent(int id)
 {
   slotStatusMsg(i18n("Opening SQL script..."));
 
-  if( !saveModified() )
-  {
-     // here saving wasn't successful
-  }
-  else
+  if( saveModified() )
   {
     // check existence of this file
     QFileInfo fi(recentFiles.at(id));
@@ -789,8 +842,16 @@ void KSqlPlusApp::slotFileSave()
 {
   slotStatusMsg(i18n("Saving SQL script..."));
 
-  if( isNew() ) slotFileSaveAs();
-  else inputView->doSave();
+  if( isNew() )
+  {
+    slotFileSaveAs();
+  }
+  else
+  {
+    inputView->doSave();
+    // update modified flag in statusbar
+    slotInputViewTextChanged();
+  }
 
   slotStatusMsg(i18n("Ready"));
 }
@@ -800,16 +861,32 @@ void KSqlPlusApp::slotFileSaveAs()
   slotStatusMsg(i18n("Saving SQL script with a new filename..."));
 
   QString file = KFileDialog::getSaveFileName(QDir::currentDirPath(),
-    i18n("*.sql|SQL scripts\n*|All files"), this, i18n("Save as..."));
+    i18n("*.sql *.SQL|SQL scripts\n*|All files"), this);
   if( !file.isEmpty() )
   {
-    inputView->doSave(file);
-    inputView->setName(file);
-    addRecentFile(file);
-    updateCaption();
-    // set path of this file as current directory
-    QFileInfo fi(file);
-    QDir::setCurrent(fi.dirPath(true));
+    // save old file name
+    QString oldfile = inputView->getName();
+    if( oldfile.isEmpty() ) oldfile = i18n("Untitled");
+    // toggle modified flag for correct saving by KEdit
+    inputView->toggleModified(true);
+    if( inputView->doSave(file) != 0 )
+    {
+      KMsgBox::message(this, i18n("Error"),
+        file + i18n(":\ncannot save SQL script to the file."),
+        KMsgBox::EXCLAMATION);
+      inputView->setName(oldfile);
+    }
+    else
+    {
+      inputView->setName(file);
+      addRecentFile(file);
+      updateCaption();
+      // set path of this file as current directory
+      QFileInfo fi(file);
+      QDir::setCurrent(fi.dirPath(true));
+    }
+    // update modified flag in statusbar
+    slotInputViewTextChanged();
   }
 
   slotStatusMsg(i18n("Ready"));
@@ -820,7 +897,7 @@ void KSqlPlusApp::slotFileSaveOutput()
   slotStatusMsg(i18n("Saving SQL*Plus output into a new file..."));
 
   QString file = KFileDialog::getSaveFileName(QDir::currentDirPath(),
-    i18n("*|All files"), this, i18n("Save output as..."));
+    i18n("*|All files"), this);
   if( !file.isEmpty() )
   {
     // open file
@@ -834,12 +911,12 @@ void KSqlPlusApp::slotFileSaveOutput()
     else
     {
       // get text buffer by line and write it to the file
-      QString s;
       for( int i = 0; i < outputView->numLines(); i++ )
       {
-        s = outputView->textLine(i);
-        s += '\n';
-        f.writeBlock(s, s.length());
+        const char* s = outputView->textLine(i);
+        if( s && *s ) f.writeBlock(s, strlen(s));
+        // write end-of-line
+        f.putch('\n');
       }
       f.flush();
     }
@@ -888,14 +965,16 @@ void KSqlPlusApp::slotFileQuit()
   slotStatusMsg(i18n("Exiting..."));
 
   saveOptions();
+
   // close the first window, the list makes the next one the first again.
   // This ensures that queryClose() is called on each window to ask for closing
-  KTMainWindow* w;
   if( memberList )
   {
+    KTMainWindow* w;
     for( w = memberList->first(); w != 0; w = memberList->first() )
     {
-      // only close the window if the closeEvent is accepted. If the user presses Cancel on the saveModified() dialog,
+      // only close the window if the closeEvent is accepted.
+      // If the user presses Cancel on the saveModified() dialog,
       // the window and the application stay open.
       if( !w->close() ) break;
     }
@@ -914,6 +993,17 @@ void KSqlPlusApp::slotEditKeys()
   slotStatusMsg(i18n("Ready"));
 }
 
+void KSqlPlusApp::slotEditSqlPlus()
+{
+  slotStatusMsg(i18n("Configuring SQL*Plus location and RSH/SSH..."));
+
+  // open SQL*Plus location dialog
+  KSqlPlusDlg dlg(this, "sqlplusDlg");
+  dlg.exec();
+
+  slotStatusMsg(i18n("Ready"));
+}
+
 void KSqlPlusApp::slotViewToolBar()
 {
   slotStatusMsg(i18n("Toggle the toolbar..."));
@@ -967,6 +1057,21 @@ void KSqlPlusApp::slotStatusHelpMsg(const char* text)
   statusBar()->message(text, 2000);
 }
 
+void KSqlPlusApp::slotClipboardChanged()
+{
+  QClipboard* clip = QApplication::clipboard();
+  if( clip && clip->text() )
+  {
+    // clipboard is filled by some text
+    enableCommand(ID_EDIT_PASTE);
+  }
+  else
+  {
+    // clipboard is empty
+    disableCommand(ID_EDIT_PASTE);
+  }
+}
+
 void KSqlPlusApp::slotInputViewCursorChanged()
 {
   // show new cursor position
@@ -987,30 +1092,19 @@ void KSqlPlusApp::slotInputViewCursorChanged()
   }
 }
 
-void KSqlPlusApp::sqlplusExecute(const char* cmd, bool addToHistory)
+void KSqlPlusApp::slotInputViewTextChanged()
 {
-  slotStatusMsg(i18n("Sending command to SQL*Plus..."));
+  // toggle the modified flag in the statusbar
+  statusBar()->changeItem((inputView->isModified() ? " * " : "   "),
+    ID_STATUS_MODIFIED);
+}
 
-  QString s(cmd);
-
-  // append command to output area
-  if( addToHistory ) appendToOutputView(s);
-
-  // send command to stdin of SQL*Plus
-  sqlplus.writeStdin(s.data(), s.length());
-
-  // append command to command history
-  if( addToHistory && commandHistory.find(s) == -1 )
-  {
-    if( commandHistory.count() >= MAX_COMMAND_HISTORY )
-      commandHistory.removeFirst();
-    commandHistory.append(s);
-    enableCommand(ID_COMMAND_HISTORY);
-    enableCommand(ID_COMMAND_PREV);
-    enableCommand(ID_COMMAND_NEXT);
-  }
-
-  slotStatusMsg(i18n("Ready"));
+void KSqlPlusApp::slotInputViewOverwriteToggled()
+{
+  // toggle the modified flag in the statusbar
+  statusBar()->changeItem(
+    (inputView->isOverwriteMode() ? i18n(" OVR ") : i18n(" INS ")),
+    ID_STATUS_OVERWRITE);
 }
 
 void KSqlPlusApp::slotCommandExecute()
@@ -1073,19 +1167,53 @@ void KSqlPlusApp::slotCommandHistory()
   slotStatusMsg(i18n("Ready"));
 }
 
-void KSqlPlusApp::slotSqlplusStdout(KProcess *proc, char* buffer, int buflen)
+void KSqlPlusApp::sqlplusExecute(const char* cmd, bool addToHistory)
+{
+  slotStatusMsg(i18n("Sending command to SQL*Plus..."));
+
+  // append command to output view
+  if( addToHistory ) appendToOutputView(cmd);
+
+  // create command and put it to the SQL*Plus command's queue
+  char* c = qstrdup(cmd);
+  sqlplusCommandQueue.enqueue(c);
+
+  // send created command to stdin of SQL*Plus
+  sqlplus.writeStdin(c, strlen(c));
+
+  // append command to command history
+  if( addToHistory && commandHistory.find(cmd) == -1 )
+  {
+    if( commandHistory.count() >= MAX_COMMAND_HISTORY )
+      commandHistory.removeFirst();
+    commandHistory.append(cmd);
+    enableCommand(ID_COMMAND_HISTORY);
+    enableCommand(ID_COMMAND_PREV);
+    enableCommand(ID_COMMAND_NEXT);
+  }
+
+  slotStatusMsg(i18n("Ready"));
+}
+
+void KSqlPlusApp::slotSqlplusWroteStdin(KProcess* proc)
+{
+  // sent command is accepted by SQL*Plus - delete it
+  sqlplusCommandQueue.remove();
+}
+
+void KSqlPlusApp::slotSqlplusStdout(KProcess* proc, char* buffer, int buflen)
 {
   // got SQL*Plus output - write it to output area
   QString s(buffer, buflen + 1);
   appendToOutputView(s);
 }
 
-void KSqlPlusApp::slotSqlplusExited(KProcess *proc)
+void KSqlPlusApp::slotSqlplusExited(KProcess* proc)
 {
   // check shutdown
   if( QApplication::closingDown() ) return;
 
-  QString msg(i18n("SQL*Plus exited"));
+  QString msg(i18n("SQL*Plus has exited"));
   if( !proc->normalExit() )
   {
     msg += i18n(" abnormally");
@@ -1112,113 +1240,87 @@ void KSqlPlusApp::commandCallback(int id)
     case ID_FILE_NEW_WINDOW:
          slotFileNewWindow();
          break;
-
     case ID_FILE_LOGIN:
     	     slotFileLogin();
          break;
-
     case ID_FILE_NEW:
     	     slotFileNew();
          break;
-
     case ID_FILE_OPEN:
          slotFileOpen();
          break;
-
     case ID_FILE_SAVE:
          slotFileSave();
          break;
-
     case ID_FILE_SAVE_AS:
          slotFileSaveAs();
          break;
-
     case ID_FILE_SAVE_OUTPUT:
          slotFileSaveOutput();
          break;
-
     case ID_FILE_CLOSE:
          slotFileClose();
          break;
-
     case ID_FILE_PRINT:
          slotFilePrint();
          break;
-
     case ID_FILE_PRINT_OUTPUT:
          slotFilePrintOutput();
          break;
-
     case ID_FILE_QUIT:
          slotFileQuit();
          break;
-
     case ID_EDIT_CUT:
          inputView->cut();
          break;
-
     case ID_EDIT_COPY:
          inputView->copyText();
          break;
-
     case ID_EDIT_PASTE:
          inputView->paste();
          break;
-  
     case ID_EDIT_CLEAR:
          inputView->clear();
          break;
-
     case ID_EDIT_COPY_OUTPUT:
          outputView->copyText();
          break;
-
     case ID_EDIT_CLEAR_OUTPUT:
          outputView->clear();
          break;
-
     case ID_EDIT_SELECT_ALL_OUTPUT:
          outputView->selectAll();
          break;
-
     case ID_EDIT_KEYS:
          slotEditKeys();
          break;
-
+    case ID_EDIT_SQLPLUS:
+         slotEditSqlPlus();
+         break;
     case ID_VIEW_TOOLBAR:
          slotViewToolBar();
          break;
-
     case ID_VIEW_STATUSBAR:
          slotViewStatusBar();
          break;
-
     case ID_VIEW_FONT:
          slotViewFont();
          break;
-
     case ID_COMMAND_EXECUTE:
          slotCommandExecute();
          break;
-
     case ID_COMMAND_BREAK:
          slotCommandBreak();
          break;
-
     case ID_COMMAND_PREV:
          slotCommandPrev();
          break;
-
     case ID_COMMAND_NEXT:
          slotCommandNext();
          break;
-
     case ID_COMMAND_HISTORY:
          slotCommandHistory();
          break;
-
-    default:
-         break;
   }
 }
 
@@ -1229,116 +1331,89 @@ void KSqlPlusApp::statusCallback(int id)
     case ID_FILE_NEW_WINDOW:
          slotStatusHelpMsg(i18n("Opens a new application window"));
          break;
-
     case ID_FILE_LOGIN:
          slotStatusHelpMsg(i18n("Connect to ORACLE"));
          break;
-
     case ID_FILE_NEW:
          slotStatusHelpMsg(i18n("Creates a new SQL script"));
          break;
-
     case ID_FILE_OPEN:
          slotStatusHelpMsg(i18n("Opens an existing SQL script"));
          break;
-
     case ID_FILE_OPEN_RECENT:
          slotStatusHelpMsg(i18n("Opens a recently used file"));
          break;
-
     case ID_FILE_SAVE:
          slotStatusHelpMsg(i18n("Saves the actual SQL script"));
          break;
-
     case ID_FILE_SAVE_AS:
          slotStatusHelpMsg(i18n("Saves the actual SQL script as..."));
          break;
-
     case ID_FILE_SAVE_OUTPUT:
          slotStatusHelpMsg(i18n("Saves the output area in file"));
          break;
-
     case ID_FILE_CLOSE:
          slotStatusHelpMsg(i18n("Closes the actual SQL script"));
          break;
-
     case ID_FILE_PRINT:
          slotStatusHelpMsg(i18n("Prints out the actual SQL script"));
          break;
-
     case ID_FILE_PRINT_OUTPUT:
          slotStatusHelpMsg(i18n("Prints out the output area"));
          break;
-
     case ID_FILE_QUIT:
          slotStatusHelpMsg(i18n("Quits the application"));
          break;
-
     case ID_EDIT_CUT:
          slotStatusHelpMsg(i18n("Cuts the selected section and puts it to the clipboard"));
          break;
-
     case ID_EDIT_COPY:
          slotStatusHelpMsg(i18n("Copies the selected section to the clipboard"));
          break;
-
     case ID_EDIT_PASTE:
          slotStatusHelpMsg(i18n("Pastes the clipboard contents to actual position"));
          break;
-
     case ID_EDIT_CLEAR:
          slotStatusHelpMsg(i18n("Clear the input area"));
          break;
-
     case ID_EDIT_COPY_OUTPUT:
          slotStatusHelpMsg(i18n("Copies the selected section in the output area to the clipboard"));
          break;
-
     case ID_EDIT_CLEAR_OUTPUT:
          slotStatusHelpMsg(i18n("Clear the output area"));
          break;
-
     case ID_EDIT_SELECT_ALL_OUTPUT:
          slotStatusHelpMsg(i18n("Select all text in the output area"));
          break;
-
     case ID_EDIT_KEYS:
          slotStatusHelpMsg(i18n("Configure application accelerators"));
          break;
-
+    case ID_EDIT_SQLPLUS:
+         slotStatusHelpMsg(i18n("Configure SQL*Plus executable location and RSH/SSH parameters"));
+         break;
     case ID_VIEW_TOOLBAR:
          slotStatusHelpMsg(i18n("Enables/disables the toolbars"));
          break;
-
     case ID_VIEW_STATUSBAR:
          slotStatusHelpMsg(i18n("Enables/disables the statusbar"));
          break;
-
     case ID_VIEW_FONT:
          slotStatusHelpMsg(i18n("Configure font for input/output views"));
          break;
-
     case ID_COMMAND_EXECUTE:
          slotStatusHelpMsg(i18n("Execute the marked or full content of input area in SQL*Plus"));
          break;
-
     case ID_COMMAND_BREAK:
          slotStatusHelpMsg(i18n("Break current command execution - kill SQL*Plus process"));
          break;
-
     case ID_COMMAND_PREV:
          slotStatusHelpMsg(i18n("Get previous command from command history into input view"));
          break;
-
     case ID_COMMAND_NEXT:
          slotStatusHelpMsg(i18n("Get next command from command history into input view"));
          break;
-
     case ID_COMMAND_HISTORY:
          slotStatusHelpMsg(i18n("Display command history dialog"));
          break;
-
-    default:
-         break;
   }
 }
